import logging

import pandas as pd

logger = logging.getLogger(__name__)


def log_dataframe_characteristics(df: pd.DataFrame, level=logging.INFO):
    """
    Given a dataframe log it's characteristics with the default logger

    Logs:
    - Dimensions
    - Column names
    - Column types

    Parameters
    ----------
    df: pd.DataFrame
        Any pandas dataframe
    level: loglevel (int) (default=logging.INFO)
        Loglevel to log at.

    Returns
    -------
    Nothing

    Examples
    --------
    >>> import pandas as pd
    >>> from sam.data_sources import read_knmi
    >>>
    >>> df = read_knmi('2018-02-01', '2019-10-01', latitude=52.11, longitude=5.18, freq='hourly',
    ...                  variables=['FH', 'FF', 'FX', 'T', 'TD', 'SQ', 'Q', 'DR', 'RH', 'P',
    ...                             'VV', 'N', 'U', 'IX', 'M', 'R', 'S', 'O', 'Y'])
    >>> from sam.logging_functions import log_dataframe_characteristics
    >>> log_dataframe_characteristics(df)
    """
    logger.log(level, f"columns: {df.shape[0]}")
    logger.log(level, f"rows: {df.shape[1]}")
    if df.shape[0] == 0 or df.shape[1] == 0:
        logger.log(level, "No type information of columns, because there were no values")
    else:
        for i in df.columns.values:
            # Take the first value of a column to get the real type
            logger.log(level, f"column: {i}, type: {type(df[i].iat[0])}")



from typing import Tuple

import numpy as np
import pandas as pd
from sam.feature_engineering import range_lag_column
from sklearn.metrics import precision_recall_curve


def incident_recall(
    y_incidents: np.ndarray,
    y_pred: np.ndarray,
    range_pred: Tuple[int, int] = (0, 0),
):
    """
    Given `y_pred`, `y_incidents` and a prediction range, see what percentage of incidents in
    `y_incidents` was positively predicted in `y_pred`, within window `range_pred`. Works for
    binary classification only (predicting 1 means incident, predicting 0 means no incident)

    For use in a make_scorer, e.g.
    `make_scorer(incident_recall, y_incidents=df['incidents'], range_pred=(1,5))`

    Parameters
    ----------
    y_incidents: 1d array-like, or label indicator array / sparse matrix
        Incidents that we want to predict with the classifier within window range_pred
    y_pred: 1d array-like, or label indicator array / sparse matrix
        Predicted labels, as returned by a classifier.
    range_pred: tuple of ints, optional (default = (0,0))
        length of two: (start prediction window, end prediction window)
        If we want to predict an incidents 5 to 1 rows in advance, this will be (1,5)
        range_pred is inclusive, so (0, 1) means you can predict either 0 or 1
        timesteps in advance

    Returns
    -------
    result: float
        The percentage of incidents that was positively predicted

    Examples
    --------
    >>> from sam.metrics import incident_recall
    >>> y_pred = [1,0,0,1,0,0,0]
    >>> y_incidents = [0,1,0,0,0,0,1]
    >>> range_pred = (0,2)
    >>> incident_recall(y_incidents, y_pred, range_pred)
    0.5
    """
    if range_pred[0] < 0 or range_pred[1] < 0:
        raise ValueError("Prediction window range_pred must be positive")
    y_pred, y_incidents = pd.Series(y_pred), pd.Series(y_incidents)
    # A prediction has effect on the future, so lag to the future
    y_pred = range_lag_column(y_pred, (-1 * range_pred[0], -1 * range_pred[1]))
    predicted_incidents = pd.concat([y_pred, y_incidents], axis=1).min(axis=1).sum()
    return predicted_incidents / y_incidents.sum()


def make_incident_recall_scorer(
    range_pred: Tuple[int, int] = (0, 0),
    colname: str = "incident",
):
    """
    Wrapper around `incident_recall_score`, to make it an actual sklearn scorer.
    This works by obtaining the 'incident' column from the data itself. This
    column name is configurable. This scorer does need the incident column
    to be present, which means by default it will be used in training the model.
    If this is not desired, a custom model will have to be made that deletes
    this variable before fitting/predicting.

    Parameters
    ----------
    range_pred: tuple of int, optional (default = (0,0))
        length of two: (start prediction window, end prediction window)
        Passed through to incident_recall
    colname: string, optional (default = "indicent")
        The column name that is given to incident_recall as incidents
        we want to predict

    Returns
    -------
    scorer: a function that acts as a sklearn scorer object.
        signature is scorer(clf, X), where clf is a fit model,
        X is test data

    Examples
    --------
    >>> from sam.metrics import make_incident_recall_scorer
    >>> from sklearn.base import BaseEstimator
    >>> import pandas as pd
    >>> import numpy as np
    >>>
    >>> op = type("MyClassifier", (BaseEstimator, object),
    ...          {"predict": lambda self, X: np.array([0, 1, 0, 0, 0, 0, 0, 0])})
    >>> data = pd.DataFrame({"incident": [0, 0, 0, 1, 1, 0, 1, 0], "other": 1})
    >>>
    >>> scorer = make_incident_recall_scorer((1, 3), "incident")
    >>> scorer(op(), data)
    0.6666666666666666
    """

    def incident_recall_scorer(clf, X):
        y_pred = clf.predict(X)
        return incident_recall(X[colname], y_pred, range_pred)

    return incident_recall_scorer


def _merge_thresholds(
    left_t: np.ndarray,
    right_t: np.ndarray,
    left_val: np.ndarray,
    right_val: np.ndarray,
):
    """
    Helper function that merges two different thresholds. Does this by iterating over the
    thresholds, and selecting the lowest threshold as the next.
    """

    def step_ahead(new_t, new_val, saved_val, ix, old_t, old_val):
        new_t.append(old_t[ix])
        new_val.append(old_val[ix])
        ix += 1
        saved_val = old_val[ix]
        return new_t, new_val, saved_val, ix

    left_ix, right_ix = 0, 0
    new_t = []
    new_leftval, new_rightval = [], []
    saved_leftval, saved_rightval = left_val[0], right_val[0]

    while left_ix < len(left_t) or right_ix < len(right_t):
        if len(left_t) > 0 and (right_ix == len(right_t) or left_t[left_ix] < right_t[right_ix]):
            new_t, new_leftval, saved_leftval, left_ix = step_ahead(
                new_t, new_leftval, saved_leftval, left_ix, left_t, left_val
            )
            new_rightval.append(saved_rightval)

        elif len(right_t) > 0 and (left_ix == len(left_t) or left_t[left_ix] > right_t[right_ix]):
            new_t, new_rightval, saved_rightval, right_ix = step_ahead(
                new_t, new_rightval, saved_rightval, right_ix, right_t, right_val
            )
            new_leftval.append(saved_leftval)

        elif left_t[left_ix] == right_t[right_ix]:
            new_t, new_leftval, saved_leftval, left_ix = step_ahead(
                new_t, new_leftval, saved_leftval, left_ix, left_t, left_val
            )
            new_t, new_rightval, saved_rightval, right_ix = step_ahead(
                new_t, new_rightval, saved_rightval, right_ix, right_t, right_val
            )
            new_t = new_t[:-1]

    new_leftval.append(left_val[-1])
    new_rightval.append(right_val[-1])
    return np.array(new_leftval), np.array(new_rightval), np.array(new_t)


def precision_incident_recall_curve(
    y_incidents: np.ndarray, y_pred: np.ndarray, range_pred: Tuple[int, int] = (0, 0)
):
    """
    Analogous to `sklearn.metrics.precision_recall_curve
    <https://scikit-learn.org/stable/modules/generated/sklearn.metrics.precision_recall_curve.html>`,
    but for incident recall and precision.
    Precision the percentage of correct incidents in the prediction.
    Incident recall can be found in the incident_recall function: for every incident, check if at
    least there is a single positive prediction.

    The calculation of the thresholds is done by calling `sklearn.precision_recall_curve`.

    Given incidents and a prediction, as well as range, returns precision, recall, thresholds.

    Parameters
    ----------
    y_incidents: 1d array-like, or label indicator array / sparse matrix
        Incidents that we want to predict with the classifier within window range_pred
    y_pred: 1d array-like, or label indicator array / sparse matrix
        Predicted labels, as returned by a classifier.
    range_pred: tuple of int, optional (default = (0,0))
        length of two: (start prediction window, end prediction window)
        If we want to predict an incidents 5 to 1 rows in advance, this will be (1,5)

    Returns
    -------
    precision: 1d array-like
        an array of precision scores
    recall: 1d array-like
        an array of incident recall scores
    thresholds: 1d array-like
        an array of thresholds to interpret the above 2

    Examples
    --------
    >>> y_pred = [0.4, 0.4,  0.1, 0.2, 0.6,  0.5, 0.1]
    >>> y_incidents = [  0, 1, 0, 0, 0, 0, 1]
    >>> range_pred = (0, 2)
    >>> p, r, t = precision_incident_recall_curve(y_incidents, y_pred, range_pred)
    >>> p
    array([0.71428571, 0.8       , 1.        , 1.        , 1.        ,
           1.        ])
    >>> r
    array([1. , 1. , 1. , 0.5, 0.5, 0. ])
    >>> t
    array([0.1, 0.2, 0.4, 0.5, 0.6])
    """
    if range_pred[0] < 0 or range_pred[1] < 0:
        raise ValueError("prediction window range_pred must be positive")

    y_lagged = range_lag_column(y_incidents, range_pred)
    precision, _, thresholds_p = precision_recall_curve(y_lagged, y_pred)

    y_pred_incidents = range_lag_column(y_pred, (-1 * range_pred[0], -1 * range_pred[1]))
    _, recall, thresholds_r = precision_recall_curve(y_incidents, y_pred_incidents)

    return _merge_thresholds(thresholds_p, thresholds_r, precision, recall)



from pathlib import Path

import pandas as pd


PACKAGEDIR = Path(__file__).parent.absolute()


def load_rainbow_beach():
    """
    Loads the Rainbow Beach dataset (subset of the open Chicago Water dataset)

    Source:
    https://data.cityofchicago.org/Parks-Recreation/Beach-Water-Quality-Automated-Sensors/qmqz-2xku
    """
    file_path = PACKAGEDIR / "data/rainbow_beach.csv"
    return pd.read_csv(file_path, index_col=[0], parse_dates=[0])


def load_sewage_data():
    """
    Loads a sewage dataset, containing the discharge of multiple pumps and some weather data

    Source: Fake dataset by Royal HaskoningDHV
    """
    file_path = PACKAGEDIR / "data/sewage_data.csv"
    return pd.read_csv(file_path, index_col=[0], parse_dates=[0])



import os
from typing import Callable, Sequence, Tuple, Union

import numpy as np
import pandas as pd
from sam.feature_engineering import BaseFeatureEngineer
from sam.models import BaseTimeseriesRegressor
from sklearn.base import TransformerMixin
from sklearn.linear_model import Lasso, QuantileRegressor
from sklearn.multioutput import MultiOutputRegressor


class LassoTimeseriesRegressor(BaseTimeseriesRegressor):
    """Linear quantile regression model (Lasso) for time series

    This model combines several approaches to time series data:
    Multiple outputs for forecasting, quantile regression, and feature engineering.
    This class implements a linear quantile regression model. For each quantile, a
    separate model is trained.

    It is a wrapper around
    the sklearn `QuantileRegressor` and `Lasso`. For more information on the
    model specifics, see the sklearn documentation.

    Parameters
    ----------
    predict_ahead: tuple of integers, optional (default=(0,))
        how many steps to predict ahead. For example, if (1, 2), the model will predict both 1 and
        2 timesteps into the future. If (0,), predict the present.
    quantiles: tuple of floats, optional (default=())
        The quantiles to predict. Values between 0 and 1. Keep in mind that the mean will be
        predicted regardless of this parameter
    use_diff_of_y: bool, optional (default=False)
        If True differencing is used (the difference between y now and shifted y),
        else differencing is not used (shifted y is used).
    timecol: string, optional (default=None)
        If not None, the column to use for constructing time features. For now,
        creating features from a DateTimeIndex is not supported yet.
    y_scaler: object, optional (default=None)
        Should be an sklearn-type transformer that has a transform and inverse_transform method.
        E.g.: StandardScaler() or PowerTransformer()
    average_type: str (default='mean')
        Determines what to fit as the average: 'mean', or 'median'. The average is the last
        node in the output layer and does not reflect a quantile, but rather estimates the central
        tendency of the data. Setting to 'mean' results in fitting that node with MSE, and
        setting this to 'median' results in fitting that node with MAE (equal to 0.5 quantile).
    feature_engineering: object, optional (default=None)
        Should be an sklearn-type transformer that has a transform method, e.g.
        `sam.feature_engineering.SimpleFeatureEngineer`.
    alpha : float, default=1.0
        Regularization constant that multiplies the L1 penalty term.
    fit_intercept : bool, default=True
        Whether or not to fit the intercept.
    quantile_options : dict, optional (default=None)
        Options for `sklearn.linear_model.QuantileRegressor`.
    mean_options : dict, optional (default=None)
        Options for `sklearn.linear_model.Lasso`.
    kwargs: dict, optional
        Not used. Just for compatibility of models that inherit from this class.

    Attributes
    ----------
    feature_engineer_: Sklearn transformer
        The transformer used on the raw data before prediction
    n_inputs_: integer
        The number of inputs used for the underlying neural network
    n_outputs_: integer
        The number of outputs (columns) from the model
    prediction_cols_: array of strings
        The names of the output columns from the model.
    model_ : object
        List of sklearn models, one for each quantile.

    Examples
    --------
    >>> import pandas as pd
    >>> from sam.models import MLPTimeseriesRegressor
    >>> from sam.feature_engineering import SimpleFeatureEngineer
    >>> from sam.datasets import load_rainbow_beach
    ...
    >>> data = load_rainbow_beach()
    >>> X, y = data, data["water_temperature"]

    >>> simple_features = SimpleFeatureEngineer(keep_original=False)
    >>> model = MLPTimeseriesRegressor(
    ...     predict_ahead=(0,),
    ...     feature_engineer=simple_features,
    ...     verbose=0,
    ... )
    >>> model.fit(X, y)  # doctest: +ELLIPSIS
    <keras.callbacks.History ...
    """

    def __init__(
        self,
        predict_ahead: Sequence[int] = (0,),
        quantiles: Sequence[float] = (),
        use_diff_of_y: bool = False,
        timecol: str = None,
        y_scaler: TransformerMixin = None,
        average_type: str = "mean",
        feature_engineer: BaseFeatureEngineer = None,
        alpha: float = 1.0,
        fit_intercept: bool = True,
        quantile_options: dict = None,
        mean_options: dict = None,
        **kwargs,
    ) -> None:
        super().__init__(
            predict_ahead=predict_ahead,
            quantiles=quantiles,
            use_diff_of_y=use_diff_of_y,
            timecol=timecol,
            y_scaler=y_scaler,
            feature_engineer=feature_engineer,
            **kwargs,
        )
        self.average_type = average_type
        self.alpha = alpha
        self.fit_intercept = fit_intercept
        self.quantile_options = quantile_options
        self.mean_options = mean_options
        self.feature_engineer = feature_engineer

        if self.average_type == "median" and 0.5 in self.quantiles:
            raise ValueError(
                "average_type is mean, but 0.5 is also in quantiles (duplicate). "
                "Either set average_type to mean or remove 0.5 from quantiles"
            )

    def get_untrained_model(self, quantile=None) -> Callable:
        """Returns linear quantile regression model"""
        if quantile is not None:
            estimator = QuantileRegressor(
                quantile=quantile,
                alpha=self.alpha,
                fit_intercept=self.fit_intercept,
                **(self.quantile_options or {}),
            )
        else:
            estimator = Lasso(
                alpha=self.alpha,
                fit_intercept=self.fit_intercept,
                **(self.mean_options or {}),
            )
        return MultiOutputRegressor(estimator=estimator)

    def fit(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        **fit_kwargs,
    ):
        X, y, _, _ = self.preprocess_fit(X, y)
        self.model_ = [self.get_untrained_model(quantile) for quantile in self.quantiles]

        if self.average_type == "mean":
            self.model_.append(self.get_untrained_model())
        elif self.average_type == "median":
            self.model_.append(self.get_untrained_model(0.5))
        else:
            raise ValueError(f"Unknown average_type: {self.average_type}")

        for model in self.model_:
            model.fit(X, y, **fit_kwargs)
        return self

    def predict(
        self,
        X: pd.DataFrame,
        y: pd.Series = None,
        return_data: bool = False,
        force_monotonic_quantiles: bool = False,
    ) -> Union[pd.DataFrame, Tuple[pd.DataFrame, pd.DataFrame]]:

        self.validate_data(X)

        if y is None and self.use_diff_of_y:
            raise ValueError("You must provide y when using use_diff_of_y=True")

        X_transformed = self.preprocess_predict(X, y)
        prediction = [model.predict(X_transformed) for model in self.model_]
        prediction = np.concatenate(prediction, axis=1)

        prediction = self.postprocess_predict(
            prediction, X, y, force_monotonic_quantiles=force_monotonic_quantiles
        )

        if return_data:
            return prediction, X_transformed
        else:
            return prediction

    def dump(self, foldername: str, prefix: str = "model") -> None:
        """Save a model to disk

        This abstract method needs to be implemented by any class inheriting from
        SamQuantileRegressor. This function dumps the SAM model to disk.

        Parameters
        ----------
        foldername : str
            The folder location where to save the model
        prefix : str, optional
           The prefix used in the filename, by default "model"
        """
        import joblib

        if not os.path.exists(foldername):
            os.makedirs(foldername)
        joblib.dump(self, os.path.join(foldername, f"{prefix}.pkl"))

    @classmethod
    def load(cls, foldername, prefix="model") -> Callable:
        """Load a model from disk

        This abstract method needs to be implemented by any class inheriting from
        SamQuantileRegressor. This function loads a SAM model from disk.

        Parameters
        ----------
        foldername : str
            The folder location where the model is stored
        prefix : str, optional
           The prefix used in the filename, by default "model"

        Returns
        -------
        The SAM model that has been loaded from disk
        """
        import joblib

        return joblib.load(os.path.join(foldername, f"{prefix}.pkl"))



import logging

import pandas as pd
from sam.logging_functions import log_new_columns

logger = logging.getLogger(__name__)


def sum_grouped_columns(df: pd.DataFrame, sep: str = "#", skipna: bool = True) -> pd.DataFrame:
    """
    Utility function to sum columns together based on groups. The column names are assumed to look
    like groupname#suffix. For example: GROUP1#lag_1_day, or GROUP2#sum_1_week. In these
    examples, the groups are GROUP1 and GROUP2 respectively. This function will find all the
    groups, and sum all the columns in the same group together. If a column does not contain the
    separator `sep`, the entire column name is assumed to be the groupname. This means columns like
    GROUP2 and GROUP2#lag_0 will be assumed to be in the same group.

    This function is mainly useful when dealing with a dataframe filled with shapley values. In
    this case, when many features are in the same group, it may be useful to sum these shapley
    values, to calculate a combined contribution that the entire group has. Keep in mind that
    this has the potential to 'wipe out' shapley values: if GROUP1#lag_0 has a large positive
    contribution, and GROUP1#lag_1 has a large negative contribution, then the group GROUP1 as
    a whole will have a contribution near 0. This is mathematically correct, and does indeed
    mean that GROUP1 as a whole had a very small effect on the prediction.

    Parameters
    ----------
    df: pd.DataFrame
       The dataframe whose columns will be added together
    sep: str, optional (default='#')
       The seperator character. The group of a column is defined as everything before the first
       occurrence of this character
    skipna: boolean, optional (default=True)
       Whether or not to ignore missing values in columns. If true, missing values are treated as
       0. Else, missing values are not ignored and the sum for that particular group/row
       combination will be missing as well.

    Returns
    -------
    summed_df: dataframe
        A dataframe with the same row-index as df, but with less columns. All the columns in the
        same group have been summed together.

    Examples
    --------
    >>> import pandas as pd
    >>> df = pd.DataFrame({
    ...    'X#lag_0': [1, 2, 3],
    ...    'X#lag_1': [1, 2, 3],
    ...    'Y': [5, 5, 5]
    ... })
    >>> sum_grouped_columns(df)
       X  Y
    0  2  5
    1  4  5
    2  6  5
    """
    logger.debug("Now running sum_grouped_columns with sep={}, skipna={}".format(sep, skipna))
    foo = df.copy()
    groups = foo.columns.str.extract(r"^([^{}]*)".format(sep)).values.reshape(-1)
    if skipna:
        result = foo.groupby(groups, axis=1).sum()
    else:
        result = foo.groupby(groups, axis=1).apply(lambda x: x.sum(skipna=False, axis=1))
    log_new_columns(result, df)
    return result


def has_strictly_increasing_index(df: pd.DataFrame, linear: bool = True) -> bool:
    """
    Utility function to validate the index of a dataframe:
    - Check if it is a DatetimeIndex
    - Check if it is strictly increasing
    - Optional: check if increases are constant over the index

    Parameters
    ----------
    df: dataframe
       The dataframe whose index needs to be checked
    linear: bool, optional (default=True)
       True to also check that the index steps are constant over the whole index length

    Returns
    -------
    bool
        Whether the index of the dataframe is a DatetimeIndex and is strictly monotonic increasing
    """

    def _index_has_constant_increase(df: pd.DataFrame):
        series_diff: pd.Series = df.index.to_series().diff()
        if series_diff.size > 1:
            # disregard first index (always 0), use uniqueness to validate the constant increase
            return False if series_diff[1:].unique().size > 1 else True
        else:
            return False

    if type(df.index) == pd.core.indexes.datetimes.DatetimeIndex and (
        df.index.is_monotonic_increasing and df.index.is_unique
    ):
        if linear and not _index_has_constant_increase(df):
            return False
        else:
            return True
    else:
        return False


def make_df_monotonic(df: pd.DataFrame, aggregate_func: str = "max") -> pd.DataFrame:
    """
    Utility function to force monotonicity over the columns of a dataframe.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe containing the columns over which we want to force monotonicity
    aggregate_func : str, optional
        Parameter to aggregate columns together.
        By default, "max" which results in a monotonic increase over the columns (left to right).
        Set to "min" to force monotonic decrease over the columns (left to right).

    Returns
    -------
    pd.DataFrame
        Dataframe that is now monotonic over the columns.

    Raises
    ------
    ValueError
        If the aggregate_func raises an exception.
    """
    if df.empty:
        return df

    df_copy = df.copy()

    _LEGAL_AGG_FUNC = ("min", "max")
    if aggregate_func not in _LEGAL_AGG_FUNC:
        raise ValueError(
            f"Illegal aggregate_func={aggregate_func}, please choose from {_LEGAL_AGG_FUNC}"
        )

    for idx, col in enumerate(df_copy.columns):
        if idx > 0:
            df_copy[col] = getattr(df_copy.iloc[:, : idx + 1], aggregate_func)(axis=1)

    return df_copy


def contains_nans(df: pd.DataFrame) -> pd.DataFrame:
    """
    Utility function to check if a dataframe contains any NaN values.

    Parameters
    ----------
    df: dataframe
       The dataframe whose index needs to be checked

    Returns
    -------
    bool
        Whether the dataframe contains any NaN values
    """
    return df.isnull().values.any()


def assert_contains_nans(df: pd.DataFrame, msg: str = "Data cannot contain nans") -> None:
    """
    Utility function to check if a dataframe contains any NaN values.

    Parameters
    ----------
    df: dataframe
        The dataframe whose index needs to be checked
    msg: str, optional (default="Data cannot contain nans")
        The error message to raise if the dataframe contains nans
    """
    if contains_nans(df):
        raise ValueError(msg)



import numpy as np
import pandas as pd


def _interpolate_pattern(bigtime, smalltime=None, pattern=0, length=1):
    """
    Helper function to create a cubic spline
    The magnitude of the cubic spline is given by pattern
    The x-values of the cubic spline are given by bigtime
    Length is the number of different values of bigtime
    For example:
    _interpolate_pattern(date.dt.month, date.dt.day, 5, 12)
    Will create a cubic spline between points on the first day of the month
    """
    if isinstance(pattern, (int, float)):
        pattern = np.random.uniform(0, pattern, length)
    else:
        for value in pattern:
            if not isinstance(value, (int, float)):
                raise TypeError("pattern must be a list of int or float")

    spline = np.full(bigtime.size, np.nan)
    for ix, value in enumerate(pattern):
        if smalltime is None:
            spline[(bigtime == ix)] = value
        else:
            spline[(bigtime == ix) & (smalltime == 0)] = value
    try:
        return pd.Series(spline).interpolate("cubic").values
    except ValueError:
        # numeric nonsense, probably happens only if the length of bigtime is too small
        return np.zeros(bigtime.size)


def _add_temporal_noise(time, noisetype="poisson", noisesize=0, length=1):
    """
    Helper function to create noise that is different but predictable
    The noise must be poisson or normal
    For example, to make some days of the week noisier than other days
    _add_temporal_noise(date.dt.month, 'normal', 2, 12)
    will generate 12 numbers between 0 and 2. Then, it will add gaussian
    noise to each month, with magnitude of the generated number.
    """
    if isinstance(noisesize, (int, float)):
        noisesize = np.random.poisson(noisesize, length)
    temp = np.zeros(time.size)
    for ix, value in enumerate(noisesize):
        if noisetype == "poisson":
            temp[time == ix] = np.random.poisson(lam=value, size=temp[time == ix].size)
        if noisetype == "normal":
            temp[time == ix] = np.random.normal(loc=0, scale=value, size=temp[time == ix].size)
    return temp


def synthetic_timeseries(
    dates,
    monthly=0,
    daily=0,
    hourly=0,
    monthnoise=(None, 0),
    daynoise=(None, 0),
    noise={},
    minmax_values=None,
    cutoff_values=None,
    negabs=None,
    random_missing=None,
    seed=None,
):
    """
    Create a synthetic time series, with some temporal patterns, and some noise. There are various
    parameters to control the distribution of the variables. The output will never be completely
    realistic, it will at least resemble what real life data could look like.

    The algorithm works like this:

    - 3 cubic splines are created: one with a monthly pattern, one with a day-of-week pattern, and
      one with an hourly pattern. These splines are added together.
    - For each month and day of the week, noise is generated according to monthnoise and daynoise
      These two sources of noise are added together
    - Noise as specified by the noise parameter is generated for each point
    - The above three series are added together. Rescale the result according to `minmax_values`
    - Missing values are added according to `cutoff_values` and `random_missing`
    - The values are mutated according to negabs

    The result is returned in a numpy array with the same length as the `dates` input.
    Due to the way the cubic splines are generated, there may be several dozen to a hundred data
    points at the beginning and end that are `nan`. To fix this, choose a dates array that is a
    couple of days longer than what you really want. Then, at the end, filter the output to only
    the dates in the middle.

    Parameters
    ----------
    dates: series of datetime, shape=(n_inputs,)
        The index of the time series that will be created. At least length 2.
        Must be a pandas series, with a `.dt` attribute.
    monthly: numeric, optional (default=0)
        The magnitude of the (random) monthly pattern. A random magnitude will be created for each
        month, with a cubic spline interpolating between months. The higher this value, the
        stronger the monthly pattern
    daily: numeric, optional (default=0)
        The magnitude of the (random) daily pattern. A random magnitude will be created for each
        day of the week, with a cubic spline interpolating between days. The higher this value,
        the stronger the daily pattern
    hourly: numeric, optional (default=0)
        The magnitude of the (random) hourly pattern. A random magnitude will be created for each
        hour, with a cubic spline interpolating between days. The higher this value, the stronger
        the daily pattern
    monthnoise: tuple of (str, numeric), optional (default=(None, 0))
        The type and magnitude of the monthly noise. For each month, a different magnitude will be
        uniformly drawn between 0 and `monthnoise[1]`. The type of the noise is given in
        `monthnoise[0]` and is either 'normal', 'poisson', or other (no noise). This noise is added
        to all points,but the magnitude wil differ between the 12 different months.
    daynoise: tuple of (str, numeric), optional (default=(None, 0))
        The type and magnitude of the daily noise. For each day of the week, a different magnitude
        will be drawn between 0 and `daynoise[1]`. The type of the noise is given in `daynoise[0]`
        and is either 'normal', 'poisson', or other (no noise). This noise is added to all points,
        but the magnitude wil differ between the 7 different days of the week.
    noise: dict, optional (default={})
        The types of noise that are added to every single point. The keys of this dictionary are
        'normal', 'poisson', or other (ignored)
        The value of the dictionary is the scale of the noise, standard deviation for normal noise,
        and the lambda value for poisson noise. The greater, the higher the variance of the result.
    minmax_values: tuple, optional (default=None)
        The values will be linearly rescaled to always fall within these bounds.
        By default, no rescaling is done.
    cutoff_values: tuple, optional (default=None)
        After rescaling, all the values that fall outside of these bounds will be set to `nan`.
        By default, no cutoff is done, and no values will be set to `nan`.
    negabs: numeric, optional (default=None)
        This value is subtracted from all the output (after rescaling), and then the result will
        be the absolute value. This oddly-specific operation is useful in case you want a positive
        value that has a lot of values around 0. This is very hard to do otherwise.
        By subtracting and taking the absolute value, this is achieved.
    random_missing: numeric, optional (default=None)
        Between 0 and 1. The fraction of values that will be set to nan. Used to emulate time
        series with a lot of missing values. The missing values will be completely randomly
        distributed with no pattern.
    seed: int or 1-d array_like, optional (default=None)
        seed for random noise generation. Passed through to `numpy.random.seed`. By default, no
        call to `numpy.random.seed` is made.

    Returns
    -------
    timeseries: numpy array, shape=(n_inputs,)
        A numpy array containing numbers, generated according to the provided parameters.

    Examples
    --------
    >>> # Create data that slightly resembles the temperature in a Nereda reactor:
    >>> from sam.data_sources.synthetic_data import synthetic_date_range, synthetic_timeseries
    >>> dates = pd.date_range('2015-01-01', '2016-01-01', freq='6H').to_series()
    >>> rnd = synthetic_timeseries(
    ...     dates,
    ...     monthly=5,
    ...     daily=1,
    ...     hourly=0.0,
    ...     monthnoise=('normal', 0.01),
    ...     daynoise=('normal', 0.01),
    ...     noise={'normal': 0.1},
    ...     minmax_values=(5, 25),
    ...     cutoff_values=None,
    ...     random_missing=0.12,
    ...     seed = 0,
    ... )
    >>> # visualize the result to see if it looks random or not
    >>> import matplotlib.pyplot as plt
    >>> fig, ax = plt.subplots()
    >>> ax = ax.plot(dates[600:700], rnd[600:700])
    >>> fig = fig.autofmt_xdate()
    >>> plt.show()  # doctest: +SKIP
    """
    if dates.size < 2:
        raise ValueError("There must be at least 2 datetimes to generate a timeseries")
    if seed is not None:
        np.random.seed(seed)
    data = np.zeros(dates.size)

    # dt.month and dt.day start at 1, the rest start at 0
    # For this reason, 1 is subtracted from month and day, because _interpolate_pattern
    # Expects its inputs to start at 1

    # add monthly pattern
    data += _interpolate_pattern(dates.dt.month - 1, dates.dt.day - 1, monthly, 12)
    # add daily pattern
    data += _interpolate_pattern(dates.dt.dayofweek, dates.dt.hour, daily, 7)
    # add hourly pattern
    data += _interpolate_pattern(dates.dt.hour, None, hourly, 24)

    # add monthly noise
    data += _add_temporal_noise(dates.dt.month - 1, monthnoise[0], monthnoise[1], 12)
    # add monthly noise
    data += _add_temporal_noise(dates.dt.dayofweek, daynoise[0], daynoise[1], 7)

    # add noise
    for key in noise:
        if key == "normal":
            data += np.random.normal(size=dates.size, loc=0, scale=noise["normal"])
        if key == "poisson":
            data += np.random.poisson(size=dates.size, lam=noise["poisson"])

    # Rescale values to all fall exactly in the minmax values
    if minmax_values is not None:
        currentmin, currentmax = np.nanmin(data), np.nanmax(data)
        scale = (minmax_values[1] - minmax_values[0]) / (currentmax - currentmin)
        offset = minmax_values[0] - currentmin
        data = scale * (data + offset)

    # Set values outside of cutoff to nan. Used for more predictable nans
    if cutoff_values is not None:
        data = np.where((data < cutoff_values[0]) | (data > cutoff_values[1]), np.nan, data)

    # Subtracts a value from the result, and absolute value.
    # This makes the result center more around 0
    if negabs is not None:
        data = np.abs(data - negabs)

    # Add random missing values, to the proportion of random_missing.
    if random_missing is not None:
        data[np.random.choice(data.size, int(data.size * random_missing), replace=False)] = np.nan

    return data


def synthetic_date_range(
    start="2016-01-01",
    end="2017-01-01",
    freq="H",
    max_delay=0,
    random_stop_freq=0,
    random_stop_max_length=1,
    seed=None,
):
    """
    Create a synthetic, somewhat realistic-looking array of datetimes.

    Given a start time, end time, frequency, and some variables governing noise,
    creates an array of datetimes that is somewhat random.

    The algorithm:

    - Generate a regular pandas date_range with start, end, and frequency
    - Delay each time by a uniformly chosen random number between 0 and `max_delay`, in seconds.
    - Pick a proportion `random_stop_freq` of times randomly. Each of these times `x_i`
      are deemed 'stoppages', and for each, a number between 1 and `random_stop_max_length` is
      uniformly chosen, say `k_i`. Then, the 'stoppage', the `k_i` next points after `x_i` are
      deleted, causing a hole in the times.
    - Only the times strictly smaller than end are kept. This means end is an exclusive bound.

    Parameters
    ----------
    start: str or datetime-like, optional (default='2016-01-01')
        Left bound for generating dates.
    end: str or datetime-like, optional (default='2017-01-01')
        Right bound for generating dates. Exclusive bound.
    freq: str or DateOffset, optional (default='H') (hourly)
        Frequency strings can have multiples, e.g. '5H'. See `here for a list of frequency aliases.
        <https://pandas.pydata.org/pandas-docs/stable/timeseries.html#timeseries-offset-aliases`_
    max_delay: numeric, optional (default=0)
        Each time is delayed by a random number of seconds, chosen between 0 and `max_delay`
    random_stop_freq: numeric, optional (default=0)
        Number between 0 and 1. This proportion of all times are deemed as starting points of
        'stoppages'. A stoppage means that a number of points are removed from the result.
    random_stop_max_length: numeric, optional (default=1)
        Each stoppage will have a randomly generated length, between 1 and
        `random_stop_max_length`. A stoppage of length `k` means that the first `k` points after
        the start of the stoppage are deleted.
    seed: int or 1-d array_like, optional (default=None)
        seed for random noise generation. Passed through to `numpy.random.seed`. By default, no
        call to `numpy.random.seed` is made.

    Returns
    -------
    rng: DatetimeIndex
        A pandas datetimeindex of noisy times

    Examples
    --------
    >>> # Generate times with point approximately every 6 hours
    >>> from sam.data_sources.synthetic_data import synthetic_date_range
    >>> synthetic_date_range('2016-01-01', '2016-01-02', '6H', 600, 0, 1, seed=0)
    DatetimeIndex(['2016-01-01 00:05:29.288102356',
                   '2016-01-01 06:12:38.401722180',
                   '2016-01-01 12:18:40.059747823',
                   '2016-01-01 18:24:06.989657621'],
                  dtype='datetime64[ns]', freq=None)

    >>> # Generate times with very likely stops of length 1
    >>> synthetic_date_range('2016-01-01', '2016-01-02', 'H', 0, 0.5, 1, seed=0)
    DatetimeIndex(['2016-01-01 00:00:00', '2016-01-01 01:00:00',
                   '2016-01-01 02:00:00', '2016-01-01 03:00:00',
                   '2016-01-01 04:00:00', '2016-01-01 05:00:00',
                   '2016-01-01 09:00:00', '2016-01-01 10:00:00',
                   '2016-01-01 11:00:00', '2016-01-01 13:00:00',
                   '2016-01-01 16:00:00', '2016-01-01 21:00:00'],
                  dtype='datetime64[ns]', freq=None)
    """
    index = pd.date_range(start, end, freq=freq).to_series()
    if index.size == 0:
        raise ValueError("End time must be after start time.")
    if seed is not None:
        np.random.seed(seed)

    # Add a delay of 0 to max_delay seconds to every time
    random_delay = np.cumsum(np.random.uniform(0, max_delay, index.size))
    index += pd.to_timedelta(random_delay, unit="s")

    # Choose indexes for the start of random stops
    random_stops = np.random.choice(len(index), int(random_stop_freq * len(index)), replace=False)
    for ix in random_stops:
        # If random_stop_max_length is 1, we cannot randomly draw a number, so just set it to 1
        stop_length = (
            1 if random_stop_max_length == 1 else np.random.randint(1, random_stop_max_length)
        )
        index[ix : ix + stop_length] = np.nan
    # Remove the stops. This causes long batches that take (much) longer than usual.
    index = index.dropna()
    # make the end an exclusive bound. It may have been surpassed because of the delays
    index = index[index < pd.to_datetime(end)]

    return pd.DatetimeIndex(index)



import logging
import warnings

import numpy as np
import pandas as pd

logger = logging.getLogger(__name__)


def label_dst(timestamps_series: pd.Series):
    """
    Find possible conflicts due to daylight savings time, by
    labeling timestamps_series. This converts a series of timestamps to
    a series of strings. The strings are either 'normal',
    'to_summertime', or 'to_wintertime'.
    to_summertime happens the last sunday morning of march,
    from 2:00 to 2:59.
    to_wintertime happens the last sunday morninig of october,
    from 2:00 to 2:59.
    These can be possible problems because they either happen 2
    or 0 times. to_summertime should therefore be impossible.

    Parameters
    ----------
    timestamps_series: pd.Series, shape = (n_inputs,)
        a series of pandas timestamps

    Returns
    -------
    labels: string, array-like, shape = (n_inputs,)
        a numpy array of strings, that are all either
        'normal', 'to_summertime', or 'to_wintertime'

    Examples
    --------
    >>> from sam.preprocessing import label_dst
    >>> import pandas as pd
    >>>
    >>> daterange = pd.date_range('2019-10-27 01:00:00', '2019-10-27 03:00:00', freq='15min')
    >>> date_labels = label_dst(pd.Series(daterange))
    >>>
    >>> pd.DataFrame({'TIME' : daterange,
    ...               'LABEL': date_labels})
                     TIME          LABEL
    0 2019-10-27 01:00:00         normal
    1 2019-10-27 01:15:00         normal
    2 2019-10-27 01:30:00         normal
    3 2019-10-27 01:45:00         normal
    4 2019-10-27 02:00:00  to_wintertime
    5 2019-10-27 02:15:00  to_wintertime
    6 2019-10-27 02:30:00  to_wintertime
    7 2019-10-27 02:45:00  to_wintertime
    8 2019-10-27 03:00:00         normal
    """
    logger.debug("Labeling dst on {} timestamps".format(timestamps_series.size))
    last_sunday_morning = (
        (timestamps_series.dt.day >= 25)
        & (timestamps_series.dt.weekday == 6)
        & (timestamps_series.dt.hour == 2)
    )
    result = np.where(
        (last_sunday_morning) & (timestamps_series.dt.month == 3),
        "to_summertime",
        np.where(
            (last_sunday_morning) & (timestamps_series.dt.month == 10),
            "to_wintertime",
            "normal",
        ),
    )

    valuecounts = str(pd.Series(result).value_counts()).replace("\n", ", ")
    logger.debug("labeldst output values: {}".format(valuecounts))
    return result


def average_winter_time(data: pd.DataFrame, tmpcol: str = "tmp_UNID"):
    """
    Solve duplicate timestamps in wintertime, by averaging them
    Because the to_wintertime hour happens twice, there can be duplicate timestamps
    This function removes those duplicates by averaging the VALUE column
    All other columns are used as group-by columns

    Parameters
    ----------
    data: pandas Dataframe
        must have columns TIME, VALUE, and optionally others like ID and TYPE.
    tmpcol: string, optional (default='tmp_UNID')
        temporary columnname that is created in dataframe. This columnname cannot
        exist in the dataframe already

    Returns
    -------
    data: pandas Dataframe
        The same dataframe as was given in input, but with duplicate timestamps
        removed, if they happened during the wintertime duplicate hour

    Examples
    --------
    >>> from sam.preprocessing import average_winter_time
    >>> import numpy as np
    >>>
    >>> daterange = pd.date_range('2019-10-27 01:45:00', '2019-10-27 03:00:00', freq='15min')
    >>> test_df = pd.DataFrame({"TIME": daterange.values[[0, 1, 1, 2, 2, 3, 3, 4, 4, 5]],
    ...                         "VALUE": np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])})
    >>> average_winter_time(test_df)
                     TIME  VALUE
    0 2019-10-27 01:45:00    0.0
    1 2019-10-27 02:00:00    1.5
    2 2019-10-27 02:15:00    3.5
    3 2019-10-27 02:30:00    5.5
    4 2019-10-27 02:45:00    7.5
    5 2019-10-27 03:00:00    9.0
    """
    warnings.warn(
        "DEPRECATED: Convert to UTC using pandas Series.dt.tz_convert() and use the"
        " `timezone`-parameter of `decompose_datetime()` to extract local time features if needed",
        DeprecationWarning,
    )

    if tmpcol in data.columns:
        raise ValueError("tmpcol already exists in dataframe")

    logging.debug("Now averaging wintertime")

    # Prevent side effects because this function makes inplace changes
    data_copy = data.copy()
    dst_labels = label_dst(data_copy.TIME)
    # We make a column that is unique for all except wintertime
    # This means that in the groupby line, non-to_wintertime
    # lines will never be grouped
    data_copy[tmpcol] = np.where(
        dst_labels == "to_wintertime", -1, np.arange(len(data_copy.index))
    )
    groupcols = data_copy.columns.tolist()
    groupcols.remove("VALUE")  # in place only
    data_copy = data_copy.groupby(groupcols).mean().reset_index().drop(tmpcol, axis=1)

    logging.info(
        "TIME colum changed because of Average_winter_time. "
        "Before it had {} rows, now it has {}".format(data.shape[0], data_copy.shape[0])
    )

    return data_copy



from typing import Callable, List

from sklearn.preprocessing import FunctionTransformer


class FunctionTransformerWithNames(FunctionTransformer):
    """
    Helper class that extends FunctionTransformer by adding 'get_feature_names'
    method. This method returns the column names of the last dataframe
    that was returned by this transformer.

    Importantly, the function that is given to this transformer must have an output x with
    attributes 'x.columns.values', such as a pandas dataframe. If the underlying function outputs a
    numpy array, this transformer will crash. In this case, it is reccommended to use the regular
    FunctionTransformer instead.

    Everything is unchanged in the FunctionTransformer, except that the default value of the
    'validate' parameter is changed to False. This mimics the behavior of sklearn from 0.22
    onwards. This is particularly important here, because validation turns pandas dataframes into
    numpy arrays, which means that the column names are removed.

    Parameters
    ----------
    func : callable, optional default=None
        The callable to use for the transformation. This will be passed
        the same arguments as transform, with args and kwargs forwarded.
        If func is None, then func will be the identity function.
    inverse_func : callable, optional default=None
        The callable to use for the inverse transformation. This will be
        passed the same arguments as inverse transform, with args and
        kwargs forwarded. If inverse_func is None, then inverse_func
        will be the identity function.
    validate : bool, optional default=False
        Indicate that the input X array should be checked before calling
        ``func``. The possibilities are:
        - If False, there is no input validation.
        - If True, then X will be converted to a 2-dimensional NumPy array or
          sparse matrix. If the conversion is not possible an exception is
          raised.
    accept_sparse : bool, optional
        Indicate that func accepts a sparse matrix as input. If validate is
        False, this has no effect. Otherwise, if accept_sparse is false,
        sparse matrix inputs will cause an exception to be raised.
    check_inverse : bool, default=True
       Whether to check that or ``func`` followed by ``inverse_func`` leads to
       the original inputs. It can be used for a sanity check, raising a
       warning when the condition is not fulfilled.
    kw_args : dict, optional
        Dictionary of additional keyword arguments to pass to func.
    inv_kw_args : dict, optional
        Dictionary of additional keyword arguments to pass to inverse_func.

    Examples
    --------
    >>> import pandas as pd
    >>> from sam.utils import FunctionTransformerWithNames
    >>> from sam.feature_engineering import decompose_datetime
    >>> data = pd.DataFrame({'TIME': [1, 2, 3], 'VALUE': [4,5,6]})
    >>> data['TIME'] = pd.to_datetime(data['TIME'])
    >>> transformer = FunctionTransformerWithNames(decompose_datetime,
    ...                                            kw_args={'components': ['hour', 'minute']})
    >>> transformer.fit_transform(data)
                               TIME  VALUE  TIME_hour  TIME_minute
    0 1970-01-01 00:00:00.000000001      4          0            0
    1 1970-01-01 00:00:00.000000002      5          0            0
    2 1970-01-01 00:00:00.000000003      6          0            0
    >>> transformer.get_feature_names_out()
    ['TIME', 'VALUE', 'TIME_hour', 'TIME_minute']
    """

    def __init__(
        self,
        func: Callable = None,
        inverse_func: Callable = None,
        validate: bool = False,
        accept_sparse: bool = False,
        check_inverse: bool = True,
        kw_args: dict = None,
        inv_kw_args: dict = None,
    ):
        super(FunctionTransformerWithNames, self).__init__(
            func=func,
            inverse_func=inverse_func,
            validate=validate,
            accept_sparse=accept_sparse,
            check_inverse=check_inverse,
            kw_args=kw_args,
            inv_kw_args=inv_kw_args,
        )

    def transform(self, X, y=None):
        """
        Applies the function, and saves the output feature names
        """
        # y is completely ignored to be consistent with all versions of sklearn
        output = super(FunctionTransformerWithNames, self).transform(X)
        self._feature_names = list(output.columns.values)
        return output

    def get_feature_names_out(self, input_features=None) -> List[str]:
        """
        Returns the feature names saved during `transform`
        """
        return self._feature_names



from typing import List, Union

import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin


class ClipTransformer(BaseEstimator, TransformerMixin):
    """
    Transformer that clips values to a given range.

    Parameters
    ----------
    cols: list (optional)
        Columns of input data to be clipped. If None, all columns will be clipped.
    min_value: float (optional)
        Minimum value to clip to. If None, min will be set to the minimum value of the data.
    max_value: float (optional)
        Maximum value to clip to. If None, max will be set to the maximum value of the data.
    """

    def __init__(
        self,
        cols: list = None,
        min_value: float = None,
        max_value: float = None,
    ):
        self.cols = cols
        self.min_value = min_value
        self.max_value = max_value

    def fit(self, X: pd.DataFrame, y: Union[pd.DataFrame, pd.Series] = None):
        """
        Fit the transformer to the data.

        Parameters
        ----------
        X: pd.DataFrame
            Dataframe containing the features to be clipped.
        y: pd.Series or pd.DataFrame (optional)
            Series or dataframe containing the target (ignored)
        """

        if self.cols is None:
            self.cols = X.columns

        if self.min_value is None:
            self.min_value = X[self.cols].min().to_dict()

        if self.max_value is None:
            self.max_value = X[self.cols].max().to_dict()

        self._feature_names_out = X.columns

        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Transform the data.

        Parameters
        ----------
        X: pd.DataFrame
            Dataframe containing the features to be clipped.
        """

        X = X.copy()
        X[self.cols] = X[self.cols].clip(self.min_value, self.max_value)
        return X

    def get_feature_names_out(self, input_features=None) -> List[str]:
        """
        Get the names of the output features.
        """

        return self._feature_names_out



import numpy as np
import pandas as pd
from sam.metrics import tilted_loss
from sklearn.base import BaseEstimator, RegressorMixin
from sam.utils.warnings import add_future_warning

# Keep package independent of statsmodels
try:
    from statsmodels.regression.quantile_regression import QuantReg
except ImportError:
    pass


class LinearQuantileRegression(BaseEstimator, RegressorMixin):
    """
    scikit-learn style wrapper for QuantReg
    Fits a linear quantile regression model, base idea from
    https://github.com/Marco-Santoni/skquantreg/blob/master/skquantreg/quantreg.py
    This class requires statsmodels

    Parameters
    ----------
    quantiles: list or float, default=[0.05, 0.95]
        Quantiles to fit, with `` 0 < q < 1 `` for each q in quantiles.
    tol: float, default=1e-3
        The tolerance for the optimization. The optimization stops
        when duality gap is smaller than the tolerance
    max_iter: int, default=1000
        The maximum number of iterations
    fit_intercept: bool, default=True
        Whether to calculate the intercept for this model. If set to false,
        no intercept will be used in calculations (e.g. data is expected to be
        already centered).

    Attributes
    ----------
    model_: statsmodel model
        The underlying statsmodel class
    model_result_: statsmodel results
        The underlying statsmodel results

    Examples
    --------
    >>> from sam.models import LinearQuantileRegression
    >>> from sam.data_sources import read_knmi
    >>> from sklearn.model_selection import train_test_split
    >>>
    >>> # Prepare data
    >>> data = read_knmi('2018-02-01', '2019-10-01', freq='hourly',
    ...                 variables=['FH', 'FF', 'FX', 'T']).set_index('TIME')
    >>> y = data['T']
    >>> X = data.drop('T', axis=1)
    >>> # Fit model
    >>> model = LinearQuantileRegression()
    >>> model.fit(X, y)  # doctest: +SKIP
    """

    @add_future_warning(
        "This class is deprecated and will be removed in a future version. "
        "Please use the LassoTimeseriesRegressor class from the sam.models "
        "or the QuantileRegressor class from sklearn.linear_model instead."
    )
    def __init__(
        self,
        quantiles: list = [0.05, 0.95],
        tol: float = 1e-3,
        max_iter: int = 1000,
        fit_intercept: bool = True,
    ):
        self.quantiles = quantiles
        self.tol = tol
        self.max_iter = max_iter
        self.fit_intercept = fit_intercept

    def _fit_single_model(self, X, y, q):
        # Statsmodels requires to add constant columns manually
        # otherwise the models will not fit an intercept
        X = X.copy()
        if self.fit_intercept:
            X = X.assign(const=1)  # f(x) = a + bX = a*const + b*X
        model_ = QuantReg(y, X)
        model_result_ = model_.fit(q, p_tol=self.tol, max_iter=self.max_iter)
        coef = model_result_.params
        pvalues = model_result_.pvalues
        return coef, pvalues

    def _array_to_df(self, X):
        """Transform to dataframe"""
        if not isinstance(X, pd.DataFrame):
            X = pd.DataFrame(X)
            X.columns = [f"X{i}" for i in range(1, X.shape[1] + 1)]
        return X

    def fit(self, X: np.array, y: np.array):
        """
        Fit a Linear Quantile Regression using statsmodels
        """
        if type(self.quantiles) is float:
            self.q_ = [self.quantiles]
        elif type(self.quantiles) is list:
            self.q_ = self.quantiles
        else:
            raise TypeError(
                f"Invalid type, quantile {self.quantiles} " f"should be a float or list of floats"
            )
        self.prediction_cols = [f"predict_q_{q}" for q in self.quantiles]

        X = self._array_to_df(X)
        self.feature_names_in_ = X.columns.tolist()
        models_ = [self._fit_single_model(X, y, q) for q in self.quantiles]
        self.coef_, self.pvalue_ = zip(*models_)
        return self

    def predict(self, X: np.array):
        """
        Predict / estimate quantiles
        """
        X = self._array_to_df(X)
        preds = [X.assign(const=1).multiply(c).sum(axis=1) for c in self.coef_]
        preds_df = pd.concat(preds, axis=1)
        preds_df.columns = self.prediction_cols
        return preds_df

    def score(self, X: np.array, y: np.array):
        """
        Default score function. Returns the tilted loss
        """
        y_pred = self.predict(X)
        scores = [
            tilted_loss(y_true=y, y_pred=y_pred[f"predict_q_{q}"], quantile=q)
            for q in self.quantiles
        ]
        score = np.mean(scores)
        return score



import logging
from typing import Union

import numpy as np
import pandas as pd
from sam.validation import BaseValidator
from sam.utils import add_future_warning

logger = logging.getLogger(__name__)


class MADValidator(BaseValidator):
    """
    This transformer finds extreme values and sets them to nan in a few steps:

    - Estimate upper and lower bounds from the data in the fit method by
      computing median deviation above and below a running median
    - Mark differences outside these bounds as nan in the transform method

    This class can be passed to the plot function (see :ref:`extreme-removal-plot`)
    to create a visualization of the removal procedure.
    It is advisory to take a look at this diagnostic plot to see if your
    `rollingwindow` parameter is sufficiently large to capture slow variations,
    without removing local peaks that might be 'outliers'.

    In addition, the default `madthresh` of 15 is relatively conservative. Less
    strict thresholds can be tried.

    Note that you only pass cols that are suited for extreme value detection.
    For instance, a pump can sometimes be out of operation and so be set to 0.
    This signal is therefore not suited for extreme value detection.

    Note that nans still have to be filled in with a later procedure.

    Note that you should fit this method to the train set!

    Parameters
    ---------
    rollingwindow: int or string
        if number, this amount of values will be used for the rolling window
        if string, should be in pandas timedelta format ('1D'), and data should
        have a datetime index. A sensible value for this depends on your time
        resolution, but you could try values between 200-400.
    cols: list of strings (optional)
        columns to detect extreme values for. If None, all columns will be used.
    madthresh: float
        number of median absolute deviations to use as threshold.

    Examples
    --------
    >>> from sam.validation import MADValidator
    >>> from sam.visualization import diagnostic_extreme_removal
    >>> import numpy as np
    >>> import pandas as pd
    >>>
    >>> # create some random data
    >>> np.random.seed(10)
    >>> data = np.random.random(size=(1000))
    >>>
    >>> # split in train and test
    >>> train_df = pd.DataFrame()
    >>> train_df['values'] = data[:800]
    >>> test_df = pd.DataFrame()
    >>> test_df['values'] = data[800:]
    >>>
    >>> # with one clear outlier
    >>> train_df.loc[25] *= 10
    >>>
    >>> # now detect extremes
    >>> cols_to_check = ['values']
    >>> REV = MADValidator(
    ...     rollingwindow=10,
    ...     cols=cols_to_check,
    ...     madthresh=10)
    >>> train_corrected = REV.fit_transform(train_df)
    >>> fig = diagnostic_extreme_removal(REV, train_df, 'values')
    >>> test_corrected = REV.transform(test_df)
    >>> fig = diagnostic_extreme_removal(REV, test_df, 'values')
    """

    def __init__(
        self,
        rollingwindow: Union[int, str],
        cols: list = None,
        madthresh=15,
    ):

        self.rollingwindow = rollingwindow
        self.cols = cols
        self.madthresh = madthresh

    def _compute_rolling(self, x: pd.Series):
        r = x.rolling(self.rollingwindow, min_periods=1, center=True).median()
        return r

    def fit(self, data: pd.DataFrame):
        """
        Estimate upper and lower bounds from the data by column by
        computing median deviation above and below a running median by column.
        This method creates the attiburte self.thresh_high and self.thresh_low
        that contain the respective bounds.

        Parameters
        ----------
        data: pd.DataFrame
            with time indices and feature columns
        """

        self.thresh_high = {}
        self.thresh_low = {}
        self.cols_ = self.cols if self.cols else data.columns.to_list()
        for c in self.cols:

            # get data
            x = data.loc[:, c]

            # determine differenc
            rolling = self._compute_rolling(x)
            diff = x.values - rolling

            # Define thresholds as number of median absolute deviations.
            # Note that as we calculate the threshold two-sided, they
            # should be computed on signed and not absolute values
            self.thresh_high[c] = np.median(diff[diff > 0]) * self.madthresh
            self.thresh_low[c] = np.median(diff[diff < 0]) * self.madthresh

        return self

    def validate(self, X: pd.DataFrame):
        """
        Sets values that fall outside bounds set in the fit method to nan

        Parameters
        ----------
        data: pd.DataFrame
            with time indices and feature columns

        Returns
        ------
        data_r: pd.DataFrame
            input data with columns marked as nan
        """

        invalid_data = pd.DataFrame(
            data=np.zeros_like(X.values).astype(bool),
            index=X.index,
            columns=X.columns,
        )

        for c in self.cols:

            # get data
            x = X.loc[:, c]

            # determine rolling and diff
            rolling = self._compute_rolling(x)
            diff = x.values - rolling

            # as thresholds are computed in signed way, we can directly compare
            extreme_value = (diff > self.thresh_high[c]) | (diff < self.thresh_low[c])

            # log number of values removed and tresholds used
            logger.info(
                f"detected {np.sum(extreme_value)} extreme values from {c}. "
                f"using upper threshold of: {round(self.thresh_high[c], 2)} "
                f"and lower threshold of: {round(self.thresh_low[c])} "
                f"using madthresh of {self.madthresh} "
                f"and rollingwindow of {str(self.rollingwindow)}"
            )

            invalid_data[c] = extreme_value

        return invalid_data


class RemoveExtremeValues(MADValidator):
    @add_future_warning("RemoveExtremeValues is deprecated. Use MADValidator instead.")
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



import pandas as pd


class MongoWrapper:
    """
    Provides a simple wrapper to the MongoDB layers

    This class provides a wrapper for basic functionality in MongoDB. We aim
    to use MongoDB as storage layer between analyses and e.g. dashboarding.

    Parameters
    ----------
    db: string
        Name of the database
    collection: string
        the name of the collection to fetch
    location: string, optional (default="localhost")
        Location of the database
    port: integer, optional (default=27017)
        Port that the database is reachable on
    **kwargs: arbitrary keyword arguments
        Passed through to `pymongo.MongoClient`

    Examples
    --------
    >>> from sam.data_sources import MongoWrapper  # doctest: +SKIP
    >>> mon = MongoWrapper('test_magweg','test_magookweg')  # doctest: +SKIP
    >>> mon.empty().add([{'test': 7}]).get()  # doctest: +SKIP
    >>> test  # doctest: +SKIP
    """

    def __init__(self, db, collection, location="localhost", port=27017, **kwargs):
        import pymongo  # Only needed now

        self.client = pymongo.MongoClient(location, port, **kwargs)
        self.db = self.client[db]
        self.collection = self.db[collection]

    def get(self, query={}, as_df=True):
        """Get as specific collection from the database

        Parameters
        ----------
        query: dictionary-like, optional (default={})
            dictionary of parameters to use in the query.
            e.g. { "address": "Park Lane 38" }
        as_df: boolean, optional (default=True)
            return the query results as a Pandas Dataframe

        Returns
        -------
        result : pandas dataframe, or list of dictionaries
            the results of the query
        """
        col = list(self.collection.find(query))

        if as_df:
            col = pd.DataFrame(col)
            col = col.drop("_id", axis=1)

        return col

    def add(self, content):
        """Get as specific collection from the database

        Parameters
        ----------
        content: list of dictionaries, or pandas dataframe
            list of items to add to the collection

        Returns
        -------
        result : self
        """
        if isinstance(content, pd.DataFrame):
            content = content.to_dict(orient="records")

        if self.collection.insert_many(content):
            return self
        else:
            return False

    def empty(self):
        """Empty the collection

        Returns
        -------
        result: self
        """
        if self.collection.delete_many({}):
            return self
        else:
            return False



import logging

import pandas as pd

logger = logging.getLogger(__name__)


def range_lag_column(
    original_column: pd.Series,
    range_shift: tuple = (0, 1),
) -> pd.Series:
    """
    Lags a column with a range. Will not lag the actual value,
    but will set a 1 in the specified range for any non-zero value.

    The range can be positive and/or negative. If negative it will 'lag'
    to the future.

    Parameters
    ----------
    original_column: pandas series
                     The original column with non-zero items to lag
    range_shift: tuple (default=(0, 1))
                 The range to lag the original column, it is inclusive.
                 A value of 0 is no lag at all.

    Returns
    -------
    pandas series
        The lagged column as a series. The input will be converted to float64.

    Example
    -------
    >>> from sam.feature_engineering import range_lag_column
    >>> import pandas as pd
    >>>
    >>> df = pd.DataFrame({"outcome" : [0, 0, 1, 0, 0, 0, 1]})
    >>> df['outcome_lag'] = range_lag_column(df['outcome'], (1, 2))
    >>> df
       outcome  outcome_lag
    0        0          1.0
    1        0          1.0
    2        1          0.0
    3        0          0.0
    4        0          1.0
    5        0          1.0
    6        1          0.0
    """
    original_column = pd.Series(original_column)
    # For loop will fail if not in order
    range_shift = sorted(range_shift)
    # Window size of the shift (+1 because inclusive)
    window_size_inclusive: int = range_shift[1] - range_shift[0] + 1

    logger.debug(
        "Now lagging range column with length: {}. Range shift: {}".format(
            original_column.size, range_shift
        )
    )

    # Reverse because we want to lag.
    # Then, we take the max which is the boolean version of 'any'
    # At the end, reverse back, which will maintain the index
    result = (
        original_column[::-1]
        .rolling(window_size_inclusive, min_periods=1)
        .max()
        .shift(range_shift[0])
        .fillna(0)[::-1]
    )

    if range_shift[0] < 0:
        result[0 : -range_shift[0]] = (
            original_column[0 : -range_shift[0]]
            .rolling(window_size_inclusive, min_periods=1)
            .max()
        )

    return result



import numpy as np
from sklearn.metrics import mean_absolute_error


def mean_absolute_scaled_error(
    y_true: np.array,
    y_pred: np.array,
    shift: int,
    sample_weight: np.array = None,
    multioutput: str = "uniform_average",
):
    """
    Given true value and predicted value, calculates MASE. Lower is better.
    MASE is the mean absolute error, divided by the MAE of a naive benchmark. The naive benchmark
    is the 'persistence benchmark': predicting the same value shift points ahead. For example,
    when `shift=1`, the naive benchmark is the MAE when for timestep `n`, we predict the value at
    timestep `n-1`. As shift increases, the naive benchmark will get worse, and the MASE will get
    better (lower).

    Even though this function does not require timestamps as input, it requires that
    the inputs are sorted by TIME (ascending), and that the timestamps are uniform. Otherwise, the
    output of this function will have no meaning.

    The shift points at the beginning of the array cannot be used for the persistence benchmark,
    but they are used for calculating the MAE of the prediction.
    If the truth vector is constant, the naive benchmark will be 0. In this case, the function
    will throw an error since MASE has no meaning.

    Parameters
    ----------
    y_true: array-like of shape = (n_samples) or (n_samples, n_outputs)
        Ground truth (correct) target values, sorted by time (ascending), with uniform timestamps
    y_pred : array-like of shape = (n_samples) or (n_samples, n_outputs)
        Estimated target values, sorted by time (ascending), with uniform timestamps
    shift: int
        The shift used when calculating the naive benchmark. For example, when the timestamps have
        a frequency of 15 minutes, and your model is predicting 1 hour ahead, then you should set
        shift to 4 to have an honest benchmark.
    sample_weight : array-like of shape = (n_samples), optional
        Sample weights.

    multioutput : string in [‘raw_values’, ‘uniform_average’]
        or array-like of shape `(n_outputs)` Defines aggregating of multiple output values.

        'raw_values':
            Returns a full set of errors in case of multioutput input.

        'uniform_average':
            Errors of all outputs are averaged with uniform weight.

    Returns
    -------
    loss: float or ndarray of floats
        MASE output is non-negative floating point. The best value is 0.0.

    Examples
    --------
    >>> y_true = [1, 2, 3]
    >>> y_pred = [0.5, 1.5, 2.5]
    >>> # persistence benchmark would have a loss of 1. Our prediction has a loss of 0.5
    >>> mean_absolute_scaled_error(y_true, y_pred, shift=1)
    0.5
    """
    if shift != int(shift) or shift <= 0:
        raise ValueError("Shift must be a positive integer")
    # So we can accept pandas/python lists as well
    y_true, y_pred = np.array(y_true), np.array(y_pred)
    # If sample_weight is not None, we have to subscript it before passing to MAE
    naive_sampleweight = sample_weight[shift:] if sample_weight is not None else None
    naive_mae = mean_absolute_error(
        y_true[shift:],
        y_true[:-shift],
        sample_weight=naive_sampleweight,
        multioutput=multioutput,
    )
    actual_mae = mean_absolute_error(
        y_true, y_pred, sample_weight=sample_weight, multioutput=multioutput
    )

    if np.isscalar(naive_mae) and naive_mae == 0:
        raise ValueError("Target vector is constant. MASE has no meaning in this circumstance.")
    elif (naive_mae == 0).any():
        raise ValueError(
            "One of the target vectors is constant. MASE has no meaning in this circumstance."
        )
    return actual_mae / naive_mae



import sys
from typing import Callable

import numpy as np
import pandas as pd
from sam.metrics import train_r2
from sklearn.metrics import mean_absolute_error


def performance_evaluation_fixed_predict_ahead(
    y_true_train: pd.Series,
    y_hat_train: pd.DataFrame,
    y_true_test: pd.Series,
    y_hat_test: pd.DataFrame,
    resolutions: list = [None],
    predict_ahead: int = 0,
    train_avg_func: Callable = np.nanmean,
    metric: str = "R2",
):
    """
    This function evaluates model performance over time for a single given predict ahead.
    It plots and returns r-squared, and creates a scatter plot of prediction vs true values.
    It does so for different temporal resolutions and for both test and train sets.
    The idea of evaluating performance for different time resolutions is that it gives
    insight into the resolution of the underlying patterns that the model is capturing
    (e.g. if the data has a very high time-resolution, the model might not capture every
    minute-to-minute change, but it could capture the hourly or daily patterns).
    Using this approach it is possible to find out what the best time-resolution is and use
    the result in, for instance, the sam.visualizations.quantile_plot.sam_quantile_plot.

    Parameters
    ----------
    y_true_train: pd.Series
        Series that contains the true train values.
    y_hat_train: pd.DataFrame
        DataFrame that contains the predicted train values (output of
        `MLPTimeseriesRegressor.predict` method)
    y_true_test: pd.Series
        Series that contains the true test values.
    y_hat_test: pd.DataFrame
        DataFrame that contains the predicted test values (output of
        `MLPTimeseriesRegressor.predict` method)
    resolutions: list (default=[None])
        List of strings (and/or None) that are interpretable by pandas resampler.
        If set to None, will return results for the native data resolution.
        Valid options are e.g.: [None], [None, '15min', '1H'], or ['1H', '1D']
    predict_ahead: int (default=0)
        Predict_ahead to display performance for
    train_avg_func: func - Callable (default=np.nanmean)
        Optional argument to pass function to calculate the train set average, by default the
        mean is used. This average is used for calculating the r2 metric.
    metric: str (default='R2')
        Optional argument to define the metric to evaluate the performance of the train and test
        set. Options:
        - 'R2': conventional r2, best used for regression
        - 'MAE': mean absolute error, best used for forecasting
        By default the 'R2' metric is used.

    Returns
    ------
    metric_df: pd.DataFrame
        Dataframe that contains the metric (r2 or mae) values per test/train set and resolution
        combination. Contains columns: ['R2' or 'MAE', 'dataset', 'resolution']
    bar_fig: matplotlib figure
        Figure object that displays the metrics for each resolution and data set.
    scatter_fig: matplotlib figure
        Figure object that displays predicted vs true data for the different resolutions.
    best_res: string
        The resolution with the maximum metric value in the train set.

    Example
    -------

    # assuming you have some y_true_train, y_true_test and predictions y_hat_train and y_hat_test:
    >>> from sam.datasets import load_rainbow_beach
    >>> from sam.feature_engineering import SimpleFeatureEngineer
    >>> from sam.models import MLPTimeseriesRegressor
    >>> from sam.visualization import performance_evaluation_fixed_predict_ahead
    ...
    >>> data = load_rainbow_beach()
    >>> X, y = data, data["water_temperature"]
    >>> test_size = int(X.shape[0] * 0.33)
    >>> train_size = X.shape[0] - test_size
    >>> X_train, y_train = X.iloc[:train_size, :], y[:train_size]
    >>> X_test, y_test = X.iloc[train_size:, :], y[train_size:]
    ...
    >>> simple_features = SimpleFeatureEngineer(
    ...     rolling_features=[
    ...         ("wave_height", "mean", 24),
    ...         ("wave_height", "mean", 12),
    ...     ],
    ...     time_features=[
    ...         ("hour_of_day", "cyclical"),
    ...     ],
    ...     keep_original=False,
    ... )
    ...
    >>> model = MLPTimeseriesRegressor(
    ...     predict_ahead=(0,1),
    ...     feature_engineer=simple_features,
    ...     verbose=0,
    ... )
    ...
    >>> model.fit(X_train, y_train)  # doctest: +ELLIPSIS
    <keras.callbacks.History ...
    >>> y_hat_train = model.predict(X_train)
    >>> y_hat_test = model.predict(X_test)
    >>> r2_df, bar_fig, scatter_fig, best_res = performance_evaluation_fixed_predict_ahead(
    ...     y_train,
    ...     y_hat_train,
    ...     y_test,
    ...     y_hat_test,
    ...     resolutions=[None, '15min', '1H', '3H', '6H', '1D'])

    >>> # display the results
    >>> bar_fig.show()
    >>> scatter_fig.show()
    >>> print('best resolution found at %s'%best_res)  # doctest: +SKIP
    >>> r2_df.head()  # doctest: +SKIP
    >>> # print some results
    >>> best_res_r2 = r2_df.loc[(r2_df['dataset']=='train') &
    ...                         (r2_df['resolution'] == best_res), 'R2'].values[0]
    >>> native_r2 = r2_df.loc[(r2_df['dataset']=='train') &
    ...                       (r2_df['resolution'] == 'native'), 'R2'].values[0]
    >>> print('best resolution found at %s (%.3f vs %.3f native)'%(
    ...         best_res, best_res_r2, native_r2))  # doctest: +SKIP
    """
    import matplotlib.pyplot as plt
    import seaborn as sns

    # initialize scatter figure
    scatter_fig = plt.figure(figsize=(len(resolutions) * 3, 6))

    # select and shift the requested predict ahead
    y_hat_train = y_hat_train[f"predict_lead_{predict_ahead}_mean"].shift(predict_ahead)
    y_hat_test = y_hat_test[f"predict_lead_{predict_ahead}_mean"].shift(predict_ahead)

    # compute the metrics for the different temporal resolutions, for train and test data
    metric_list, dataset_list, resolution_list = [], [], []
    for ri, res in enumerate(resolutions):

        # resample data to desired resolution of requested
        if res is not None:
            res_label = res
            y_true_train_res = y_true_train.resample(res).mean()
            y_hat_train_res = y_hat_train.resample(res).mean()
            y_true_test_res = y_true_test.resample(res).mean()
            y_hat_test_res = y_hat_test.resample(res).mean()
        else:
            res_label = "native"
            y_true_train_res = y_true_train
            y_hat_train_res = y_hat_train
            y_true_test_res = y_true_test
            y_hat_test_res = y_hat_test

        # train set performance
        metric_list = _evaluate_performance(
            y_true_train_res,
            y_hat_train_res,
            y_true_train_res,
            train_avg_func,
            metric,
            metric_list,
        )

        # test set performance
        metric_list = _evaluate_performance(
            y_true_test_res,
            y_hat_test_res,
            y_true_train_res,
            train_avg_func,
            metric,
            metric_list,
        )

        # append results to lists
        dataset_list.append("train")
        resolution_list.append(res_label)
        dataset_list.append("test")
        resolution_list.append(res_label)

        # create scatter plot of train results:
        alpha = np.min([1000 / len(y_true_train_res), 1])
        plt.subplot(2, len(resolutions), ri + 1)
        ymin = np.min([y_true_train.min(), y_hat_train.min()])
        ymax = np.max([y_true_train.max(), y_hat_train.max()])
        plt.plot([ymin, ymax], [ymin, ymax], c="gray", ls="--")
        plt.plot(y_true_train_res.values, y_hat_train_res.values, "o", alpha=alpha)
        plt.title("train " + res_label)
        plt.xlim(ymin, ymax)
        plt.ylim(ymin, ymax)
        if ri > 0:
            plt.xticks([])
            plt.yticks([])
        else:
            plt.xlabel("true")
            plt.ylabel("predicted")

        # create scatter plot of test results:
        plt.subplot(2, len(resolutions), ri + 1 + len(resolutions))
        plt.plot([ymin, ymax], [ymin, ymax], c="gray", ls="--")
        plt.plot(
            y_true_test_res.values,
            y_hat_test_res.values,
            "o",
            alpha=alpha,
            color="orange",
        )
        plt.title("test " + res_label)
        plt.xlim(ymin, ymax)
        plt.ylim(ymin, ymax)
        if ri > 0:
            plt.xticks([])
            plt.yticks([])
        else:
            plt.xlabel("true")
            plt.ylabel("predicted")

    # options for scatter plot
    sns.despine()
    plt.tight_layout()

    # create bar plot of different metrics
    metric_df = pd.DataFrame(
        {metric: metric_list, "dataset": dataset_list, "resolution": resolution_list}
    )
    bar_fig = plt.figure(figsize=(6, 4))
    plt.axhline(0, c="k")
    sns.barplot(data=metric_df, x="resolution", y=metric, hue="dataset")
    if metric == "R2":
        plt.ylabel("Variance Explained (%)")
        plt.ylim(0, 100)
    elif metric == "MAE":
        plt.ylabel("Mean Absolute Error (mm)")
    sns.despine()

    # calculate best resolution as the optimal resolution metric in the train set
    if metric == "R2":
        best_res = metric_df.iloc[metric_df.loc[metric_df["dataset"] == "train", metric].idxmax()][
            "resolution"
        ]
    elif metric == "MAE":
        best_res = metric_df.iloc[metric_df.loc[metric_df["dataset"] == "train", metric].idxmin()][
            "resolution"
        ]

    return metric_df, bar_fig, scatter_fig, best_res


def _evaluate_performance(
    y_true: pd.Series,
    y_hat: pd.Series,
    y_benchmark: pd.Series = None,
    avg_func: Callable = np.nanmean,
    metric: str = "R2",
    metric_list: list = None,
):
    """
    This function evaluates model performance using the specified metric.

    Parameters
    ----------
    y_true: pd.Series
        Series that contains the true values.
    y_hat: pd.Series
        Series that contains the predicted values.
    y_benchmark: pd.Series (default=None)
        Series that serves as a benchmark for r2 evaluation, required for calculating the r2 metric
    avg_func: func - Callable (default=np.nanmean)
        Optional argument to pass function to calculate the y_benchmark average, by default the
        mean is used. This average function is required for calculating the r2 metric.
    metric: str (default='R2')
        Optional argument to define the metric to evaluate the performance of the train and test
        set. Options:
        - 'R2': conventional r2, best used for regression
        - 'MAE': mean absolute error, best used for forecasting
        By default the 'R2' metric is used.
    metric_list: list (default=None)
        Optional argument to define a list that is appended with the performance results.

    Returns
    ------
    metric_list: list
        List with performance result
    """

    if metric_list is None:
        metric_list = []

    if metric == "R2":
        if y_benchmark is None:
            raise ValueError("y_benchmark needs to be supplied")
        else:
            # compute r2 with custom r2 function (in sam.metrics)
            try:
                benchmark = avg_func(y_benchmark)
            except Exception:
                e = sys.exc_info()[0]
                raise ValueError(f"Supplied avg_func resulted in error: {e}")

        r2 = train_r2(y_true, y_hat, benchmark)
        metric_list.append(r2 * 100)
    elif metric == "MAE":
        illegal_idx = y_true.isin([np.nan, -np.inf, np.inf]) | y_hat.isin(
            [np.nan, -np.inf, np.inf]
        )
        mae = mean_absolute_error(y_true[~illegal_idx], y_hat[~illegal_idx])
        metric_list.append(mae)
    else:
        raise ValueError(f"Unknown metric '{metric}'")

    return metric_list



from typing import List, Tuple, Union, Optional

import numpy as np
import pytz
import pandas as pd
from sam.feature_engineering import BaseFeatureEngineer


class SimpleFeatureEngineer(BaseFeatureEngineer):
    """
    Base class for simple time series feature engineering. Provides a method to
    create two types of features: rolling features and time components (one hot or cyclical).

    Parameters
    ----------
    rolling_features : list or pandas.DataFrame (default=[])
        List of tuples of the form (column, method, window). Can also be provided as a
        dataframe with columns: ['column', 'method', 'window'].
        The column is the name of the column to be transformed, the method is the
        method to be used (string), and the window is the window size (integer or string).
        Valid methods are "lag" or any of the pandas rolling methods (e.g. "mean", "median", etc.).
    time_features : list (default=[])
        List of tuples of the form (component, type). Can also be provided as a
        dataframe with columns ['component', 'type'].
        For supported component values, see `SimpleFeatureEngineer.valid_components`
        (e.g. "second_of_day", "hour_of_day). Valid types are "onehot" or "cyclical".
    time_col : str (default=None)
        Name of the time column (e.g. "TIME"). If None, the index of the dataframe is used.
    timezone: str, optional (default=None)
        if tz is not None, convert the time to the specified timezone, before creating features.
        timezone can be any string that is recognized by pytz, for example `Europe/Amsterdam`.
        We assume that the TIME column is always in UTC,
        even if the datetime object has no tz info.
    drop_first : bool (default=True)
        Whether to drop the first value of time components (used for onehot encoding)
    keep_original : bool (default=False)
        Whether to keep the original columns in the dataframe.

    Example
    -------
    >>> from sam.feature_engineering import SimpleFeatureEngineer
    >>> from sam.data_sources import read_knmi
    >>> data = read_knmi('2018-02-01', '2019-10-01', latitude=52.11, longitude=5.18, freq='hourly',
    ...                  variables=['FH', 'FF', 'FX', 'T', 'TD', 'SQ', 'Q', 'DR', 'RH', 'P',
    ...                             'VV', 'N', 'U', 'IX', 'M', 'R', 'S', 'O', 'Y'])
    >>> y = data['T'].shift(-1)
    >>> X = data.copy()
    >>> fe = SimpleFeatureEngineer(rolling_features=[('T', 'mean', 12), ('T', 'mean', 24)],
    ...                            time_features=[('hour_of_week', 'onehot')], time_col="TIME")
    >>> X_fe = fe.fit_transform(X, y)
    >>> X_fe.head()  # doctest: +SKIP
    """

    component_range = {
        "second_of_minute": (0, 59),
        "second_of_hour": (0, 3599),
        "second_of_day": (0, 86399),
        "minute_of_hour": (0, 59),
        "minute_of_day": (0, 1439),
        "hour_of_day": (0, 23),
        "hour_of_week": (0, 167),
        "day_of_week": (1, 7),
        "day_of_month": (1, 31),
        "day_of_year": (1, 366),
        "week_of_year": (1, 53),
        "month_of_year": (1, 12),
    }

    component_function = {
        "second_of_minute": lambda x: x.dt.second,
        "second_of_hour": lambda x: x.dt.second + x.dt.minute * 60,
        "second_of_day": lambda x: x.dt.second + x.dt.minute * 60 + x.dt.hour * 3600,
        "minute_of_hour": lambda x: x.dt.minute,
        "minute_of_day": lambda x: x.dt.minute + x.dt.hour * 60,
        "hour_of_day": lambda x: x.dt.hour,
        "hour_of_week": lambda x: x.dt.hour + x.dt.weekday * 24,
        "day_of_week": lambda x: x.dt.isocalendar().day,
        "day_of_month": lambda x: x.dt.day,
        "day_of_year": lambda x: x.dt.dayofyear,
        "week_of_year": lambda x: x.dt.isocalendar().week,
        "month_of_year": lambda x: x.dt.month,
    }

    valid_components = component_range.keys()

    def __init__(
        self,
        rolling_features: Optional[Union[List[Tuple], pd.DataFrame]] = None,
        time_features: Optional[Union[List[Tuple], pd.DataFrame]] = None,
        time_col: Optional[str] = None,
        timezone: Optional[str] = None,
        drop_first: bool = True,
        keep_original: bool = False,
    ) -> None:
        super().__init__()
        self.rolling_features = self._input_df_to_list(rolling_features)
        self.time_features = self._input_df_to_list(time_features)
        self.time_col = time_col
        self.timezone = timezone
        self.drop_first = drop_first
        self.keep_original = keep_original

    @staticmethod
    def _input_df_to_list(
        data: pd.DataFrame,
    ) -> List[Tuple]:
        """Convert a dataframe to a list of tuples."""
        if data is None:
            return []
        if isinstance(data, pd.DataFrame):
            return list(data.to_records(index=False))
        elif isinstance(data, list):
            return data
        else:
            raise ValueError(f"Invalid data type: {type(data)}, provide a list or dataframe")

    def _fix_timezone(self, datetime):
        """Change timezone before calculating components."""
        if self.timezone is not None:
            if datetime.dt.tz is not None:
                if datetime.dt.tz != pytz.utc:
                    raise ValueError(
                        "Data should either be in UTC timezone or it should have no"
                        " timezone information (assumed to be in UTC)"
                    )
            else:
                datetime = datetime.dt.tz_localize("UTC")
            datetime = datetime.dt.tz_convert(self.timezone)
        return datetime

    def _get_time_column(self, X: pd.DataFrame, component: str) -> pd.Series:
        """Get the time column."""
        # First select the datetime column
        if self.time_col is not None:
            datetime = X[self.time_col]
        else:
            datetime = X.index.to_series().copy()

        # Fix timezone
        datetime = self._fix_timezone(datetime)

        # Then select the component
        if component in self.valid_components:
            return self.component_function[component](datetime)
        else:
            raise ValueError(f"Invalid component: {component}")

    def _get_rolling_features(self, X: pd.DataFrame) -> pd.DataFrame:
        """Get the rolling features."""
        X_out = pd.DataFrame(index=X.index, columns=[])
        # Rolling features
        if self.time_col is not None:
            X = X.set_index(self.time_col)

        for col, method, window in self.rolling_features:
            colname = f"{col}_{method}_{window}"
            if method == "lag":
                X_out[colname] = X[col].shift(window).values
            else:
                X_out[colname] = X[col].rolling(window=window).agg(method).values

        if self.time_col is not None:
            X = X.reset_index(drop=False)
        return X_out

    def _get_time_features(self, X: pd.DataFrame) -> pd.DataFrame:
        """Get the time features."""
        X_out = pd.DataFrame(index=X.index, columns=[])

        for component, type in self.time_features:
            colname = f"{component}_{type}"
            comp_min, comp_max = self.component_range[component]

            if type == "onehot":
                # we do not make a dummy of the last value because of collinearity
                if self.drop_first:
                    comp_min += 1
                for value in range(comp_min, comp_max + 1):
                    comp_series = self._get_time_column(X, component)
                    colname_ = f"{colname}_{value}"
                    X_out[colname_] = (comp_series == value).astype(int)

            elif type == "cyclical":
                comp_series = self._get_time_column(X, component)
                # scale to 0,1, then to 0,2pi and then to -1,1
                comp_norm = (comp_series - comp_min) / (comp_max - comp_min + 1)
                X_out[colname + "_sin"] = np.sin(2 * np.pi * comp_norm).astype(float)
                X_out[colname + "_cos"] = np.cos(2 * np.pi * comp_norm).astype(float)
            else:
                raise ValueError(f"Invalid type: {type}")

        return X_out

    def feature_engineer_(self, X: pd.DataFrame) -> pd.DataFrame:
        """Feature engineering function that creates rolling features and time components."""
        X = X.copy()
        if self.keep_original:
            X_out = X.copy()
        else:
            X_out = pd.DataFrame(index=X.index, columns=[])

        X_out = pd.concat([X_out, self._get_rolling_features(X)], axis=1)
        X_out = pd.concat([X_out, self._get_time_features(X)], axis=1)

        return X_out



from .correct_extremes import (
    correct_above_threshold,
    correct_below_threshold,
    correct_outside_range,
)
from .data_scaling import scale_train_test
from .differencing import (
    inverse_differenced_target,
    make_differenced_target,
    make_shifted_target,
)
from .clip_transformer import ClipTransformer
from .normalize_timestamps import normalize_timestamps
from .rnn_reshape import RecurrentReshaper
from .sam_reshape import sam_format_to_wide, wide_to_sam_format
from .time import average_winter_time, label_dst
from .train_test_split import datetime_train_test_split

__all__ = [
    "ClipTransformer",
    "normalize_timestamps",
    "correct_above_threshold",
    "correct_below_threshold",
    "correct_outside_range",
    "label_dst",
    "average_winter_time",
    "sam_format_to_wide",
    "wide_to_sam_format",
    "scale_train_test",
    "RecurrentReshaper",
    "make_shifted_target",
    "make_differenced_target",
    "inverse_differenced_target",
    "datetime_train_test_split",
]



import logging

import numpy as np
import pandas as pd
from sam.logging_functions import log_dataframe_characteristics

logger = logging.getLogger(__name__)


def incident_curves(
    data: pd.DataFrame,
    under_conf_interval: bool = True,
    max_gap: int = 0,
    min_duration: int = 0,
    max_gap_perc: float = 1,
    min_dist_total: float = 0,
    actual: str = "ACTUAL",
    low: str = "PREDICT_LOW",
    high: str = "PREDICT_HIGH",
):
    """
    Finds and labels connected outliers, or 'curves'. The basic idea of this function is to define
    an outlier as a row where the value is outside some interval. 'Interval' here refers to a
    prediction interval, that can be different for every row. This interval defines what a model
    considers a 'normal' value for that datapoint. Missing values are not considered outliers.

    Then, we apply various checks and filters to the outlier to create 'curves', or streaks of
    connected outlies. These curves can have gaps, if `max_gap > 0`. In the end, only the curves
    that satisfy conditions are kept. Curves that do not satisfy one of the conditions are ignored
    (essentially, the output will act as if they are not outliers at all).

    The output is an array of the same length as the number of rows as data, with each streak of
    outliers labeled with an unique number.
    This algorithm assumes the input is sorted by time, adjacent rows are adjacent measurements!

    Parameters
    ----------
    data: pd.DataFrame (n_rows, _)
        dataframe containing values, as well two columns determining an interval. These three
        column names can be configured using the 'actual', 'low', and 'high' parameters
    under_conf_interval: bool, optional (default=True)
        If true, values lower than the interval count as outliers. Else, only
        values higher than the interval are counted as outliers.
    max_gap: int, optional (default=0)
        How many gaps are allowed between outliers. For example, if max_gap = 2, and the outliers
        look like: [True, False, False, True], then this 'gap' of 2 will be assigned to this
        curve, and this will turn into a single curve of outliers, of length 4
    min_duration: int, optional (default=0)
        Minimum number of outliers per curve. Curves with a smaller length than this value,
        will be ignored. Gaps are counted in the duration of a curve.
    max_gap_perc: float, optional (default=1)
        The maximum percentage of gaps that a curve can have to count. For example,
        if this is 0.4, and a curve contains 2 outliers and 2 gaps, then the curve will be
        ignored.
    min_dist_total: float, optional (default=0)
        If a curve should have a minimum 'outlier size', and if so, how much.
        The outlier size here is defined as the distance between the value and the end of
        the interval. For example, if the interval is (10, 20) and the value is 21, the
        'outlier size' is 1. These values are summed (gaps are counted as 0), and compared
        to this value. Curves with a sum that is too low will be ignored.
    actual: string, optional (default='ACTUAL')
        The name of the column in the data containing the value for each row
    low: string, optional (default='PREDICT_LOW')
        The name of the column in the data containing the lower end of the interval for each row
    high: string, optional (default='PREDICT_HIGH')
        The name of the column in the data containing the higher end of the interval for each row

    Returns
    -------
    outlier_curves: array-like (n_rows,)
        A numpy array of numbers labeling each curve. 0 means there is no outlier curve
        here that satisfies all conditions.

    Examples
    --------
    >>> from sam.exploration import incident_curves
    >>> data = pd.DataFrame({'ACTUAL': [0.3, np.nan, 0.3, np.nan, 0.3, 0.5, np.nan, 0.7],
    ...                      'PREDICT_HIGH': 0.6, 'PREDICT_LOW': 0.4})
    >>> incident_curves(data)
    array([1, 0, 2, 0, 3, 0, 0, 4])
    >>> incident_curves(data, max_gap=1)
    array([1, 1, 1, 1, 1, 0, 0, 2])
    >>> incident_curves(data, max_gap=1, max_gap_perc=0.2)
    array([0, 0, 0, 0, 0, 0, 0, 2])
    """

    def _number_true_streaks(series):
        """Given a boolean series, numbers series of True and set False to 0. For example,
        [T, T, F, T, F, T, T] will be numbered as [1, 1, 0, 2, 0, 3, 3]. Each streak of T gets
        the same number. The result will be converted to numpy array.
        """
        begin_streak = series.ne(series.shift())
        # We only want beginning of True streaks, not False
        begin_true_streak = series & begin_streak
        return np.where(series, begin_true_streak.cumsum(), 0)

    logger.debug(
        "Finding outlier curves: under_conf_interval={}, max_gap={}, min_duration={}, "
        "max_gap_perc={}, min_dist_total={}, actual={}, low={}, high={}".format(
            under_conf_interval,
            max_gap,
            min_duration,
            max_gap_perc,
            min_dist_total,
            actual,
            low,
            high,
        )
    )

    data = data.copy()  # copy because we are going to mess with this df in various ways
    data = data.rename(columns={actual: "ACTUAL", low: "PREDICT_LOW", high: "PREDICT_HIGH"})
    # Three columns needed: ACTUAL, PREDICT_HIGH, PREDICT_LOW
    data["OUTLIER"] = (
        (data.ACTUAL > data.PREDICT_HIGH) | (data.ACTUAL < data.PREDICT_LOW)
        if under_conf_interval
        else (data.ACTUAL > data.PREDICT_HIGH)
    )
    # Find the streaks of gaps. Gaps are defined as anything that's not an outlier
    data["GAP"] = _number_true_streaks(~data["OUTLIER"])
    # Then, all gaps with length of max_gap or lower are merged with neighbouring outliers
    new_val = data.groupby("GAP").apply(lambda x: True if (x.shape[0] <= max_gap) else False)
    new_val[0] = True  # because this is not an outlier, it's treated seperately
    new_val.name = "OUTLIER_FILLED"  # Attribute needed for join
    data = data.join(new_val, on="GAP")  # Add OUTLIER_FILLED column

    # If there is an outlier AND a gap at the beginning/end. That is wrong.
    firstgap_id, lastgap_id = data.GAP.iloc[0], data.GAP.iloc[-1]
    if firstgap_id != 0:  # entire gap at the beginning that shouldn't be there
        data.loc[data.GAP == firstgap_id, "OUTLIER_FILLED"] = False
    if lastgap_id != 0:  # entire gap at the end that shouldn't be there
        data.loc[data.GAP == lastgap_id, "OUTLIER_FILLED"] = False

    # Calculate OUTLIER_DIST which is needed to interpret min_dist_total parameter
    data["OUTLIER_DIST"] = np.where(
        data.OUTLIER,
        np.where(
            data.ACTUAL > data.PREDICT_HIGH,
            data.ACTUAL - data.PREDICT_HIGH,
            data.PREDICT_LOW - data.ACTUAL if under_conf_interval else 0,
        ),
        0,
    )

    # Now calculate the streaks of outliers, and number them
    data["OUTLIER_CURVE"] = _number_true_streaks(data["OUTLIER_FILLED"])

    logging.debug("Curves found after fixing gaps: {}".format(data["OUTLIER_CURVE"].max()))

    # Lastly, filter the outlier streaks that don't match one of the three criteria from params.
    real_outlier = data.groupby("OUTLIER_CURVE").apply(
        lambda x: (x.shape[0] >= min_duration)
        and (x.OUTLIER_DIST.sum() >= min_dist_total)
        and (1 - (x.OUTLIER.sum() / x.shape[0]) <= max_gap_perc)
    )
    real_outlier.name = "REAL_OUTLIER"
    data = data.join(real_outlier, on="OUTLIER_CURVE")

    logging.info("Curves found in final result: {}".format(data["OUTLIER_CURVE"].max()))
    return np.where(data.REAL_OUTLIER, data.OUTLIER_CURVE, 0)


def incident_curves_information(
    data: pd.DataFrame,
    under_conf_interval: bool = True,
    return_aggregated: bool = True,
    normal: str = "PREDICT",
    time: str = "TIME",
    **kwargs,
):
    """
    Aggregates a dataframe by incident_curves.
    This function calculates incident_curves using incident_curves, and then calculates
    information about each outlier curve. This function can either return raw information
    about each outlier row, keeping the number of rows the same, or it can aggregate the dataframe
    by outlier curve, which means there will be specific information per curve, such as the length,
    the total outlier distance, etcetera.

    The data must contain the actual, low, high values (actual is the actual value, and low/high
    are some interval deeming what is 'normal') Also, the data must contain a 'time' column.
    Lastly, the data must contain a 'normal' column, describing what would be considered
    the most normal value (for example the middle of the interval)
    This algorithm assumes the input is sorted by time, adjacent rows are adjacent measurements!

    Parameters
    ----------
    data: pd.DataFrame (n_rows, _)
        dataframe containing actual, low, high, normal, and time columns. These column names
        can be configured using those parameters. The default values are (ACTUAL, PREDICT_LOW,
        PREDICT_HIGH, PREDICT, TIME)
    under_conf_interval: bool, optional (default=True)
        If true, values lower than the interval count as outliers. Else, only
        values higher than the interval are counted as outliers.
    return_aggregated: bool, optional (default=True)
        If true the information about the outliers will be aggregated by OUTLIER_CURVE.
        Else, information will not be aggregated. The two options return different
        types of information.
    normal: string, optional (default='PREDICT')
        The name of the column in the data containing a 'normal' value for each row
    time: string, optional (default='TIME')
        The name of the column in the data containing a 'time' value for each row

    Returns
    -------
    information, dataframe
        if return_aggregated is false: information about each outlier.
        The output will have the folowing columns:

        - all the original columns
        - OUTLIER (bool) whether the value of the row is considered an outlier
        - OUTLIER_CURVE (numeric) the streak the outlier belongs to, or 0 if it's
          not an outlier
        - OUTLIER_DIST (numeric) The distance between the value and the outside of
          the interval, describing how much 'out of the normal' the value is
        - OUTLIER_SCORE (numeric) If x is OUTLIER_DIST, and y is the distance
          between the value and the 'normal' column, then this is x / (1 + y)
          This defines some 'ratio' of how abnormal the value is. This can be useful
          in scale-free data, where the absolute distance is not a fair metric.
        - OUTLIER_TYPE (string) 'positive' if the outlier is above the interval,
          and 'negative' if the outlier is below the interval

        if `return_aggregated`, then it will return information about each outlier curve
        The output will have the following columns:

        - index: OUTLIER_CURVE (numeric) The id of the curve. 0 is not included
        - OUTLIER_DURATION (numeric) The number of points in the curve, including gaps
        - OUTLIER_TYPE (string) if the first point is positive or negative. Other points
          in the curve may have other types
        - OUTLIER_SCORE_MAX (numeric) The maximum of OUTLIER_SCORE of all the points
          in the curve
        - OUTLIER_START_TIME (datetime) The value of the 'time' column of the first point
          in the curve
        - OUTLIER_END_TIME (datetime) The value of the 'time' column of the last point
          in the curve
        - OUTLIER_DIST_SUM (numeric) The sum of OUTLIER_DIST of the points in the curve.
          Gaps count as 0
        - OUTLIER_DIST_MAX (numeric) The max of OUTLIER_DIST of the points in the curve

    Examples
    --------
    >>> data = pd.DataFrame({'TIME': range(1547477436, 1547477436+3),  # unix timestamps
    ...                     'ACTUAL': [0.3, 0.5, 0.7],
    ...                     'PREDICT_HIGH': 0.6, 'PREDICT_LOW': 0.4, 'PREDICT': 0.5})
    >>> incident_curves_information(data)  # doctest: +ELLIPSIS
                   OUTLIER_DURATION  ...
    >>> incident_curves_information(data, return_aggregated=False)
             TIME  ACTUAL  PREDICT_HIGH  ...  OUTLIER_DIST  OUTLIER_SCORE  OUTLIER_TYPE
    0  1547477436     0.3           0.6  ...           0.1       0.090909      negative
    1  1547477437     0.5           0.6  ...           0.0       0.000000          none
    2  1547477438     0.7           0.6  ...           0.1       0.090909      positive
    <BLANKLINE>
    [3 rows x 10 columns]
    """
    data = data.copy()
    data = data.rename(columns={normal: "PREDICT", time: "TIME"})
    logging.debug("Creating outlier information: return_aggregated={}".format(return_aggregated))
    data["OUTLIER_CURVE"] = incident_curves(data, under_conf_interval, **kwargs).astype(np.int64)
    # On unix, 64 is already the default, but on windows, the
    # outlier_curves function returns 32 bit integers. This function returns a numpy array
    # which is consistent across platforms, so values are converted to 64-bit to ensure
    # consistency.

    data["OUTLIER"] = (
        (data.ACTUAL > data.PREDICT_HIGH) | (data.ACTUAL < data.PREDICT_LOW)
        if under_conf_interval
        else (data.ACTUAL > data.PREDICT_HIGH)
    )
    data["OUTLIER_DIST"] = np.where(
        data.OUTLIER,
        np.where(
            data.ACTUAL > data.PREDICT_HIGH,
            data.ACTUAL - data.PREDICT_HIGH,
            data.PREDICT_LOW - data.ACTUAL if under_conf_interval else 0,
        ),
        0,
    )
    data["OUTLIER_SCORE"] = np.where(
        data.OUTLIER,
        np.where(
            data.ACTUAL > data.PREDICT_HIGH,
            (data.ACTUAL - data.PREDICT_HIGH) / (1 + data.PREDICT_HIGH - data.PREDICT),
            (data.PREDICT_LOW - data.ACTUAL) / (1 + data.PREDICT - data.PREDICT_LOW)
            if under_conf_interval
            else 0,
        ),
        0,
    )
    data["OUTLIER_TYPE"] = np.where(
        data.ACTUAL > data.PREDICT_HIGH,
        "positive",
        np.where(data.ACTUAL < data.PREDICT_LOW, "negative", "none"),
    )

    if not return_aggregated:
        return data

    streaks = data.groupby("OUTLIER_CURVE").agg(
        {
            "OUTLIER_CURVE": "count",
            "OUTLIER_DIST": ["sum", "max"],
            "OUTLIER_SCORE": ["max"],
            "TIME": ["min", "max"],
            "OUTLIER_TYPE": lambda x: x.iloc[0],
        }
    )
    streaks.columns = ["_".join(x) for x in streaks.columns]
    streaks = streaks.rename(
        columns={
            "OUTLIER_CURVE_count": "OUTLIER_DURATION",
            "TIME_min": "OUTLIER_START_TIME",
            "TIME_max": "OUTLIER_END_TIME",
            "OUTLIER_SCORE_max": "OUTLIER_SCORE_MAX",
            "OUTLIER_DIST_sum": "OUTLIER_DIST_SUM",
            "OUTLIER_DIST_max": "OUTLIER_DIST_MAX",
            "OUTLIER_TYPE_<lambda>": "OUTLIER_TYPE",
        }
    )
    # reorder columns
    streaks = streaks[
        [
            "OUTLIER_DURATION",
            "OUTLIER_TYPE",
            "OUTLIER_SCORE_MAX",
            "OUTLIER_START_TIME",
            "OUTLIER_END_TIME",
            "OUTLIER_DIST_SUM",
            "OUTLIER_DIST_MAX",
        ]
    ]
    logger.info("Created data from incident_curves_information:")
    log_dataframe_characteristics(streaks, logging.INFO)
    return streaks[streaks.index != 0]



from .mongo_wrapper import MongoWrapper
from .synthetic_data import synthetic_date_range, synthetic_timeseries
from .weather.knmi import read_knmi, read_knmi_station_data, read_knmi_stations
from .weather.openweathermap import read_openweathermap
from .weather.regenradar import read_regenradar

__all__ = [
    "synthetic_timeseries",
    "synthetic_date_range",
    "read_knmi",
    "read_knmi_station_data",
    "read_knmi_stations",
    "read_openweathermap",
    "read_regenradar",
    "MongoWrapper",
]



from .find_incidents import incident_curves, incident_curves_information
from .lag_correlation import lag_correlation
from .top_correlation import top_n_correlations, top_score_correlations

__all__ = [
    "incident_curves",
    "incident_curves_information",
    "lag_correlation",
    "top_n_correlations",
    "top_score_correlations",
]
